/**
 * @Author: Gongxh
 * @Date: 2025-03-20
 * @Description: 热更新管理器
 */

import { game, native, sys } from "cc";
import { Platform } from "../global/Platform";
import { ICheckUpdatePromiseResult, IPromiseResult } from "../interface/PromiseResult";
import { ReadNetFile } from "../net/nettools/ReadNetFile";
import { debug, log } from "../tool/log";
import { Utils } from "../tool/Utils";

interface IHotUpdateConfig {
    packageUrl: string;
    remoteManifestUrl: string;
    remoteVersionUrl: string;
}

export enum HotUpdateCode {
    /** 成功 */
    SUCCEED = 0,
    /** 重复初始化 */
    REPEAT_INIT = -1001,
    /** 平台不支持 不需要热更新 */
    PLATFORM_NOT_SUPPORTED = -1002,
    /** 未初始化 */
    NOT_INITIALIZED = -1003,
    /** 更新中 */
    UPDATING = -1004,

    /** 加载本地manifest失败 */
    LOAD_LOCAL_MANIFEST_FAILED = -1005,
    /** 下载manifest文件失败 */
    DOWNLOAD_MANIFEST_FAILED = -1006,
    /** 解析manifest文件失败 */
    PARSE_MANIFEST_FAILED = -1007,

    /** 更新失败 需要重试 */
    UPDATE_FAILED = -1008,
    /** 更新错误 */
    UPDATE_ERROR = -1009,
    /** 解压错误 */
    DECOMPRESS_ERROR = -1010,

    /** 下载 hotconfig 文件失败 */
    DOWNLOAD_HOTCONFIG_FAILED = -1011,

    /** 解析 hotconfig 文件失败 */
    PARSE_HOTCONFIG_FAILED = -1012,
}

const TAG = "hotupdate:";

export class HotUpdateManager {
    private static instance: HotUpdateManager;
    public static getInstance(): HotUpdateManager {
        if (!HotUpdateManager.instance) {
            HotUpdateManager.instance = new HotUpdateManager();
        }
        return HotUpdateManager.instance;
    }
    /** 本地manifest路径 */
    private _manifestUrl: string = '';
    /** 热更新配置路径 (可选) 用来动态替换manifest文件中关联的地址 */
    private _hotconfigUrl: string = '';
    /** 版本号 */
    private _version: string = '';
    /** 可写路径 */
    private _writablePath: string = '';
    /** 资源管理器 */
    private _am: native.AssetsManager = null;
    /** 是否正在更新 或者 正在检查更新 */
    private _updating: boolean = false;

    /** 更新回调 */
    private _updateProgress: (kb: number, total: number) => void = null;    // 更新进度
    private _updateFail: (res: IPromiseResult) => void = null;      // 更新失败
    private _finish: (res: IPromiseResult) => void = null;          // 更新结束

    /** 
     * 获取资源版本号, 须初始化成功后再使用
     * @return 资源版本号 默认值 ‘0’
     */
    public get resVersion(): string {
        return this._am?.getLocalManifest()?.getVersion() || "0";
    }

    /**
     * 1. 初始化热更新管理器
     * @param manifestUrl 传入本地manifest文件地址 资源的assets.nativeUrl
     * @param version 传入游戏版本号 eg: 1.0.0
     * @param hotconfigUrl 传入热更新配置地址 用于替换manifest文件中关联的地址
     */
    public init(manifestUrl: string, version: string, hotconfigUrl?: string): Promise<IPromiseResult> {
        return new Promise((resolve, reject) => {
            if (!Platform.isNativeMobile) {
                reject({ code: HotUpdateCode.PLATFORM_NOT_SUPPORTED, message: "当前平台不支持热更新" });
            }
            if (this._am) {
                reject({ code: HotUpdateCode.REPEAT_INIT, message: "请勿重复初始化" });
            }
            this._manifestUrl = manifestUrl;
            this._hotconfigUrl = hotconfigUrl || '';
            this._version = version;

            let writablePath = native?.fileUtils?.getWritablePath() || "";
            if (!writablePath.endsWith("/")) {
                writablePath += "/";
            }
            this._writablePath = `${writablePath}hot-update/${version}/`;
            log(`${TAG}可写路径:${this._writablePath}`);

            // 创建 am 对象
            this._am = native.AssetsManager.create("", this._writablePath);
            this._am.setVersionCompareHandle(Utils.compareVersion);
            this._am.setVerifyCallback(this._verifyCallback);
            // 加载本地的 manifest
            log(`${TAG} 加载本地的 manifest:${manifestUrl}`);
            if (this._am.loadLocalManifest(manifestUrl)) {
                resolve({ code: HotUpdateCode.SUCCEED, message: "初始化成功" });
            } else {
                reject({ code: HotUpdateCode.LOAD_LOCAL_MANIFEST_FAILED, message: "加载本地manifest失败" });
            }
        });
    }

    /** 加载远程的hotconfig文件 并替换local manifest内容 */
    private loadRemoteHotConfig(): Promise<IPromiseResult> {
        return new Promise((resolve, reject) => {
            new ReadNetFile({
                url: this._hotconfigUrl,
                timeout: 5,
                responseType: "text",
                onComplete: (data: string) => {
                    log(`${TAG} 下载hotconfig文件成功`);
                    if (Utils.isJsonString(data)) {
                        let hotconfig: IHotUpdateConfig = JSON.parse(data);
                        let content = native.fileUtils.getStringFromFile(this._manifestUrl); // 这个一定有
                        // 使用拿到的cdn上的配置，修改本地manifest内容
                        let newManifest: IHotUpdateConfig = JSON.parse(content);

                        newManifest.remoteManifestUrl = hotconfig.remoteManifestUrl;
                        newManifest.remoteVersionUrl = hotconfig.remoteVersionUrl;
                        newManifest.packageUrl = hotconfig.packageUrl;
                        // 注册本地manifest根目录
                        let manifestRoot = "";
                        let found = this._manifestUrl.lastIndexOf("/");
                        if (found === -1) {
                            found = this._manifestUrl.lastIndexOf("\\");
                        }
                        if (found !== -1) {
                            manifestRoot = this._manifestUrl.substring(0, found + 1);
                        }
                        this._am.getLocalManifest().parseJSONString(JSON.stringify(newManifest), manifestRoot);
                        // log(TAG + "manifest root:" + this._am.getLocalManifest().getManifestRoot());
                        // log(TAG + "manifest packageUrl:" + this._am.getLocalManifest().getPackageUrl());
                        // log(TAG + "manifest version:" + this._am.getLocalManifest().getVersion());
                        // log(TAG + "manifest versionFileUrl:" + this._am.getLocalManifest().getVersionFileUrl());
                        // log(TAG + "manifest manifestFileUrl:" + this._am.getLocalManifest().getManifestFileUrl());
                        // 重新加载本地manifest
                        resolve({ code: HotUpdateCode.SUCCEED, message: "更新热更新配置成功" });
                    } else {
                        log(`${TAG} hotconfig.manifest文件格式错误`);
                        resolve({ code: HotUpdateCode.PARSE_HOTCONFIG_FAILED, message: "hotconfig.manifest文件格式错误" });
                    }
                },
                onError: (code: number, message: string) => {
                    log(`${TAG} 下载hotconfig.manifest文件失败`, code, message);
                    // 下载hotconfig.manifest文件失败
                    resolve({ code: HotUpdateCode.DOWNLOAD_HOTCONFIG_FAILED, message: "下载hotconfig.manifest文件失败" });
                }
            });
        })
    }

    /** 
     * 检查是否存在热更新
     * 提供一个对外的方法检查是否存在热更新
     * @return {Promise<ICheckUpdatePromiseResult>} 
     */
    public checkUpdate(): Promise<ICheckUpdatePromiseResult> {
        return new Promise((resolve, reject) => {
            if (!Platform.isNativeMobile) {
                reject({ code: HotUpdateCode.PLATFORM_NOT_SUPPORTED, message: "当前平台不需要热更新" });
                return;
            }
            if (!this._am) {
                reject({ code: HotUpdateCode.NOT_INITIALIZED, message: "未初始化, 需要先调用init方法" });
                return;
            }
            if (this._updating) {
                reject({ code: HotUpdateCode.UPDATING, message: "正在更新或者正在检查更新" });
                return;
            }
            this._updating = true;
            // 设置回调
            this._am.setEventCallback((event: native.EventAssetsManager) => {
                let eventCode = event.getEventCode();
                log(`${TAG} 检查更新回调code:${eventCode}`);
                this._updating = false;
                switch (eventCode) {
                    case native.EventAssetsManager.ERROR_DOWNLOAD_MANIFEST:
                        this._am.setEventCallback(null);
                        this._updating = false;
                        resolve({ code: HotUpdateCode.DOWNLOAD_MANIFEST_FAILED, message: "下载manifest文件失败", needUpdate: false, size: 0 });
                        return;
                    case native.EventAssetsManager.ERROR_PARSE_MANIFEST:
                        this._am.setEventCallback(null);
                        this._updating = false;
                        resolve({ code: HotUpdateCode.PARSE_MANIFEST_FAILED, message: "解析manifest文件失败", needUpdate: false, size: 0 });
                        return;
                    case native.EventAssetsManager.ALREADY_UP_TO_DATE:
                        this._am.setEventCallback(null);
                        this._updating = false;
                        resolve({ code: HotUpdateCode.SUCCEED, message: "已是最新版本", needUpdate: false, size: 0 });
                        return;
                    case native.EventAssetsManager.NEW_VERSION_FOUND:
                        // 发现新版本
                        this._am.setEventCallback(null);
                        this._updating = false;
                        resolve({ code: HotUpdateCode.SUCCEED, message: "发现新版本", needUpdate: true, size: this._am.getTotalBytes() / 1024 });
                        return;
                }
            });

            // 如果存在替换地址
            if (this._hotconfigUrl) {
                log(`${TAG} 下载 hotconfig 文件`);
                this.loadRemoteHotConfig().then((res) => {
                    if (res.code === HotUpdateCode.SUCCEED || res.code === HotUpdateCode.DOWNLOAD_HOTCONFIG_FAILED) {
                        this._am.checkUpdate();
                    } else {
                        this._updating = false;
                        reject(res);
                    }
                });
            } else {
                // 使用本地的menifest检查更新
                log(`${TAG} 使用本地的menifest检查更新`);
                this._am.checkUpdate();
            }
        });
    }

    /**
     * 开始热更新
     * @param res.progress 更新进度回调 kb: 已下载的资源大小, total: 总资源大小 (kb)
     * @param res.fail 更新失败了 等待重试
     * @param res.finish 更新结束了 根据code 判断是否跳过
     */
    public startUpdate(res: {
        progress: (kb: number, total: number) => void,
        fail: (res: IPromiseResult) => void,
        finish: (res: IPromiseResult) => void
    }): void {
        if (!Platform.isNativeMobile) {
            res.finish({ code: HotUpdateCode.PLATFORM_NOT_SUPPORTED, message: "当前平台不需要热更新" });
        }
        if (!this._am) {
            res.finish({ code: HotUpdateCode.NOT_INITIALIZED, message: "未初始化, 需要先调用init方法" });
            return;
        }
        if (this._updating) {
            res.finish({ code: HotUpdateCode.UPDATING, message: "正在更新或者正在检查更新" });
            return;
        }
        this._updateProgress = res.progress;
        this._updateFail = res.fail;
        this._finish = res.finish;

        log(`${TAG} 开始热更新`);
        this.checkUpdate().then((result) => {
            if (result.needUpdate) {
                this.update();
            } else {
                res.finish({ code: HotUpdateCode.SUCCEED, message: "已是最新版本" });
            }
        }).catch((err) => {
            res.finish({ code: err.code, message: err.message });
        });
    }

    private update(): void {
        this._updating = true;
        this._am.setEventCallback((event: native.EventAssetsManager) => {
            let eventCode = event.getEventCode();
            debug(`${TAG} 更新回调code:${eventCode}`);
            switch (eventCode) {
                case native.EventAssetsManager.UPDATE_PROGRESSION: {
                    let bytes = event.getDownloadedBytes() / 1024;
                    let total = event.getTotalBytes() / 1024;
                    this._updateProgress(bytes, total);
                    break;
                }
                case native.EventAssetsManager.UPDATE_FINISHED: {
                    this._updating = false;
                    // 更新完成 自动重启
                    this._am.setEventCallback(null);

                    // Prepend the manifest's search path
                    let searchPaths = native.fileUtils.getSearchPaths();
                    log(`${TAG} 当前搜索路径:${JSON.stringify(searchPaths)}`);

                    let newPaths = this._am.getLocalManifest().getSearchPaths();
                    log(`${TAG} 新搜索路径:${JSON.stringify(newPaths)}`);

                    Array.prototype.unshift.apply(searchPaths, newPaths);
                    sys.localStorage.setItem('hotupdate::version', this._version);
                    sys.localStorage.setItem('hotupdate::searchpaths', JSON.stringify(searchPaths));
                    native.fileUtils.setSearchPaths(searchPaths);

                    // 重启游戏
                    setTimeout(() => {
                        game.restart();
                    }, 500);
                    break;
                }
                case native.EventAssetsManager.UPDATE_FAILED: {
                    this._updating = false;
                    this._updateFail({ code: HotUpdateCode.UPDATE_FAILED, message: event.getMessage() });
                    break;
                }
                case native.EventAssetsManager.ERROR_UPDATING: {
                    this._updating = false;
                    this._finish({ code: HotUpdateCode.UPDATE_ERROR, message: event.getMessage() });
                    break;
                }
                case native.EventAssetsManager.ERROR_DECOMPRESS: {
                    this._updating = false;
                    this._finish({ code: HotUpdateCode.DECOMPRESS_ERROR, message: event.getMessage() });
                    break;
                }
                default:
                    break;
            }
        });
        this._am.update();
    }

    /** 重试失败的资源 */
    public retryUpdate(): void {
        this._am.downloadFailedAssets();
    }

    /** 验证资源 */
    private _verifyCallback(path: string, asset: native.ManifestAsset): boolean {
        // 资源是否被压缩, 如果压缩我们不需要检查它的md5值
        let compressed = asset.compressed;
        if (compressed) {
            return true;
        }
        // 预期的md5
        let expectedMD5 = asset.md5;
        // 资源大小
        let size = asset.size;
        // 验证资源md5
        log(`${TAG} 记录的md5:${expectedMD5} 文件大小:${size} 文件相对路径:${asset.path} 绝对路径:${path}`);
        return true;
    }
}
